@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: ogr_core.h 33680 2016-03-08 09:59:03Z rouault $
+ * $Id: ogr_core.h cf7ac1b48089a3ef06e2f4d514300968c7b36c47 2021-04-02 19:42:56 +0200 Even Rouault $
  *
  * Project:  OpenGIS Simple Features Reference Implementation
  * Purpose:  Define some core portability services for cross-platform OGR code.
@@ -7,7 +7,7 @@
  *
  ******************************************************************************
  * Copyright (c) 1999, Frank Warmerdam
- * Copyright (c) 2007-2014, Even Rouault <even dot rouault at mines-paris dot org>
+ * Copyright (c) 2007-2014, Even Rouault <even dot rouault at spatialys.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -32,6 +32,9 @@
 #define OGR_CORE_H_INCLUDED
 
 #include "cpl_port.h"
+#if defined(GDAL_COMPILATION)
+#define DO_NOT_DEFINE_GDAL_RELEASE_DATE_AND_GDAL_RELEASE_NAME
+#endif
 #include "gdal_version.h"
 
 /**
@@ -40,69 +43,77 @@
  * Core portability services for cross-platform OGR code.
  */
 
-/**
- * Simple container for a bounding region.
- */
-
 #if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)
+
+extern "C++"
+{
+#if !defined(DOXYGEN_SKIP)
+#include <limits>
+#endif
+
+/**
+ * Simple container for a bounding region (rectangle)
+ */
 class CPL_DLL OGREnvelope
 {
   public:
-        OGREnvelope() : MinX(0.0), MaxX(0.0), MinY(0.0), MaxY(0.0)
+        /** Default constructor. Defines an empty rectangle  */
+        OGREnvelope() : MinX(std::numeric_limits<double>::infinity()),
+                        MaxX(-std::numeric_limits<double>::infinity()),
+                        MinY(std::numeric_limits<double>::infinity()),
+                        MaxY(-std::numeric_limits<double>::infinity())
         {
         }
 
+        /** Copy constructor */
         OGREnvelope(const OGREnvelope& oOther) :
             MinX(oOther.MinX),MaxX(oOther.MaxX), MinY(oOther.MinY), MaxY(oOther.MaxY)
         {
         }
 
+        /** Assignment operator */
+        OGREnvelope& operator=(const OGREnvelope&) = default;
+
+    /** Minimum X value */
     double      MinX;
+
+    /** Maximum X value */
     double      MaxX;
+
+    /** Minimum Y value */
     double      MinY;
+
+    /** Maximum Y value */
     double      MaxY;
 
 #ifdef HAVE_GCC_DIAGNOSTIC_PUSH
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wfloat-equal"
 #endif
-    int  IsInit() const { return MinX != 0 || MinY != 0 || MaxX != 0 || MaxY != 0; }
+    /** Return whether the object has been initialized, that is, is non empty */
+    int  IsInit() const { return MinX != std::numeric_limits<double>::infinity(); }
 
 #ifdef HAVE_GCC_DIAGNOSTIC_PUSH
 #pragma GCC diagnostic pop
 #endif
 
+    /** Update the current object by computing its union with the other rectangle */
     void Merge( OGREnvelope const& sOther ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,sOther.MinX);
-            MaxX = MAX(MaxX,sOther.MaxX);
-            MinY = MIN(MinY,sOther.MinY);
-            MaxY = MAX(MaxY,sOther.MaxY);
-        }
-        else
-        {
-            MinX = sOther.MinX;
-            MaxX = sOther.MaxX;
-            MinY = sOther.MinY;
-            MaxY = sOther.MaxY;
-        }
-    }
-    void Merge( double dfX, double dfY ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,dfX);
-            MaxX = MAX(MaxX,dfX);
-            MinY = MIN(MinY,dfY);
-            MaxY = MAX(MaxY,dfY);
-        }
-        else
-        {
-            MinX = MaxX = dfX;
-            MinY = MaxY = dfY;
-        }
+        MinX = MIN(MinX,sOther.MinX);
+        MaxX = MAX(MaxX,sOther.MaxX);
+        MinY = MIN(MinY,sOther.MinY);
+        MaxY = MAX(MaxY,sOther.MaxY);
     }
 
+    /** Update the current object by computing its union with the provided point */
+    void Merge( double dfX, double dfY ) {
+        MinX = MIN(MinX,dfX);
+        MaxX = MAX(MaxX,dfX);
+        MinY = MIN(MinY,dfY);
+        MaxY = MAX(MaxY,dfY);
+    }
+
+    /** Update the current object by computing its intersection with the other rectangle */
     void Intersect( OGREnvelope const& sOther ) {
         if(Intersects(sOther))
         {
@@ -123,25 +134,27 @@ class CPL_DLL OGREnvelope
         }
         else
         {
-            MinX = 0;
-            MaxX = 0;
-            MinY = 0;
-            MaxY = 0;
+            *this = OGREnvelope();
         }
     }
 
+    /** Return whether the current object intersects with the other rectangle */
     int Intersects(OGREnvelope const& other) const
     {
         return MinX <= other.MaxX && MaxX >= other.MinX &&
                MinY <= other.MaxY && MaxY >= other.MinY;
     }
 
+    /** Return whether the current object contains the other rectangle */
     int Contains(OGREnvelope const& other) const
     {
         return MinX <= other.MinX && MinY <= other.MinY &&
                MaxX >= other.MaxX && MaxY >= other.MaxY;
     }
 };
+
+} // extern "C++"
+
 #else
 typedef struct
 {
@@ -152,75 +165,70 @@ typedef struct
 } OGREnvelope;
 #endif
 
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)
+
+extern "C++" {
 
 /**
  * Simple container for a bounding region in 3D.
  */
-
-#if defined(__cplusplus) && !defined(CPL_SURESS_CPLUSPLUS)
 class CPL_DLL OGREnvelope3D : public OGREnvelope
 {
   public:
-        OGREnvelope3D() : OGREnvelope(), MinZ(0.0), MaxZ(0.0)
+        /** Default constructor. Defines an empty rectangle  */
+        OGREnvelope3D() : OGREnvelope(),
+                          MinZ(std::numeric_limits<double>::infinity()),
+                          MaxZ(-std::numeric_limits<double>::infinity())
         {
         }
 
+        /** Copy constructor */
         OGREnvelope3D(const OGREnvelope3D& oOther) :
                             OGREnvelope(oOther),
                             MinZ(oOther.MinZ), MaxZ(oOther.MaxZ)
         {
         }
 
+        /** Assignment operator */
+        OGREnvelope3D& operator=(const OGREnvelope3D&) = default;
+
+    /** Minimum Z value */
     double      MinZ;
+
+    /** Maximum Z value */
     double      MaxZ;
 
 #ifdef HAVE_GCC_DIAGNOSTIC_PUSH
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wfloat-equal"
 #endif
-    int  IsInit() const { return MinX != 0 || MinY != 0 || MaxX != 0 || MaxY != 0 || MinZ != 0 || MaxZ != 0; }
+    /** Return whether the object has been initialized, that is, is non empty */
+    int  IsInit() const { return MinX != std::numeric_limits<double>::infinity(); }
 #ifdef HAVE_GCC_DIAGNOSTIC_PUSH
 #pragma GCC diagnostic pop
 #endif
 
+    /** Update the current object by computing its union with the other rectangle */
     void Merge( OGREnvelope3D const& sOther ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,sOther.MinX);
-            MaxX = MAX(MaxX,sOther.MaxX);
-            MinY = MIN(MinY,sOther.MinY);
-            MaxY = MAX(MaxY,sOther.MaxY);
-            MinZ = MIN(MinZ,sOther.MinZ);
-            MaxZ = MAX(MaxZ,sOther.MaxZ);
-        }
-        else
-        {
-            MinX = sOther.MinX;
-            MaxX = sOther.MaxX;
-            MinY = sOther.MinY;
-            MaxY = sOther.MaxY;
-            MinZ = sOther.MinZ;
-            MaxZ = sOther.MaxZ;
-        }
-    }
-    void Merge( double dfX, double dfY, double dfZ ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,dfX);
-            MaxX = MAX(MaxX,dfX);
-            MinY = MIN(MinY,dfY);
-            MaxY = MAX(MaxY,dfY);
-            MinZ = MIN(MinZ,dfZ);
-            MaxZ = MAX(MaxZ,dfZ);
-        }
-        else
-        {
-            MinX = MaxX = dfX;
-            MinY = MaxY = dfY;
-            MinZ = MaxZ = dfZ;
-        }
+        MinX = MIN(MinX,sOther.MinX);
+        MaxX = MAX(MaxX,sOther.MaxX);
+        MinY = MIN(MinY,sOther.MinY);
+        MaxY = MAX(MaxY,sOther.MaxY);
+        MinZ = MIN(MinZ,sOther.MinZ);
+        MaxZ = MAX(MaxZ,sOther.MaxZ);
     }
 
+    /** Update the current object by computing its union with the provided point */
+    void Merge( double dfX, double dfY, double dfZ ) {
+        MinX = MIN(MinX,dfX);
+        MaxX = MAX(MaxX,dfX);
+        MinY = MIN(MinY,dfY);
+        MaxY = MAX(MaxY,dfY);
+        MinZ = MIN(MinZ,dfZ);
+        MaxZ = MAX(MaxZ,dfZ);
+    }
+
+    /** Update the current object by computing its intersection with the other rectangle */
     void Intersect( OGREnvelope3D const& sOther ) {
         if(Intersects(sOther))
         {
@@ -245,15 +253,11 @@ class CPL_DLL OGREnvelope3D : public OGREnvelope
         }
         else
         {
-            MinX = 0;
-            MaxX = 0;
-            MinY = 0;
-            MaxY = 0;
-            MinZ = 0;
-            MaxZ = 0;
+            *this = OGREnvelope3D();
         }
     }
 
+    /** Return whether the current object intersects with the other rectangle */
     int Intersects(OGREnvelope3D const& other) const
     {
         return MinX <= other.MaxX && MaxX >= other.MinX &&
@@ -261,6 +265,7 @@ class CPL_DLL OGREnvelope3D : public OGREnvelope
                MinZ <= other.MaxZ && MaxZ >= other.MinZ;
     }
 
+    /** Return whether the current object contains the other rectangle */
     int Contains(OGREnvelope3D const& other) const
     {
         return MinX <= other.MinX && MinY <= other.MinY &&
@@ -268,6 +273,9 @@ class CPL_DLL OGREnvelope3D : public OGREnvelope
                MinZ <= other.MinZ && MaxZ >= other.MaxZ;
     }
 };
+
+} // extern "C++"
+
 #else
 typedef struct
 {
@@ -280,45 +288,49 @@ typedef struct
 } OGREnvelope3D;
 #endif
 
-
 CPL_C_START
 
-void CPL_DLL *OGRMalloc( size_t );
-void CPL_DLL *OGRCalloc( size_t, size_t );
-void CPL_DLL *OGRRealloc( void *, size_t );
-char CPL_DLL *OGRStrdup( const char * );
-void CPL_DLL OGRFree( void * );
+/*! @cond Doxygen_Suppress */
+void CPL_DLL *OGRMalloc( size_t ) CPL_WARN_DEPRECATED("Use CPLMalloc instead.");
+void CPL_DLL *OGRCalloc( size_t, size_t ) CPL_WARN_DEPRECATED("Use CPLCalloc instead.");
+void CPL_DLL *OGRRealloc( void *, size_t ) CPL_WARN_DEPRECATED("Use CPLRealloc instead.");
+char CPL_DLL *OGRStrdup( const char * ) CPL_WARN_DEPRECATED("Use CPLStrdup instead.");
+void CPL_DLL OGRFree( void * ) CPL_WARN_DEPRECATED("Use CPLFree instead.");
+/*! @endcond */
 
 #ifdef STRICT_OGRERR_TYPE
+/** Type for a OGR error */
 typedef enum
 {
-    OGRERR_NONE,
-    OGRERR_NOT_ENOUGH_DATA,
-    OGRERR_NOT_ENOUGH_MEMORY,
-    OGRERR_UNSUPPORTED_GEOMETRY_TYPE,
-    OGRERR_UNSUPPORTED_OPERATION,
-    OGRERR_CORRUPT_DATA,
-    OGRERR_FAILURE,
-    OGRERR_UNSUPPORTED_SRS,
-    OGRERR_INVALID_HANDLE,
-    OGRERR_NON_EXISTING_FEATURE
+    OGRERR_NONE,                       /**< Success */
+    OGRERR_NOT_ENOUGH_DATA,            /**< Not enough data to deserialize */
+    OGRERR_NOT_ENOUGH_MEMORY,          /**< Not enough memory */
+    OGRERR_UNSUPPORTED_GEOMETRY_TYPE,  /**< Unsupported geometry type */
+    OGRERR_UNSUPPORTED_OPERATION,      /**< Unsupported operation */
+    OGRERR_CORRUPT_DATA,               /**< Corrupt data */
+    OGRERR_FAILURE,                    /**< Failure */
+    OGRERR_UNSUPPORTED_SRS,            /**< Unsupported SRS */
+    OGRERR_INVALID_HANDLE,             /**< Invalid handle */
+    OGRERR_NON_EXISTING_FEATURE        /**< Non existing feature. Added in GDAL 2.0 */
 } OGRErr;
 #else
+/** Type for a OGR error */
 typedef int OGRErr;
 
-#define OGRERR_NONE                0
-#define OGRERR_NOT_ENOUGH_DATA     1    /* not enough data to deserialize */
-#define OGRERR_NOT_ENOUGH_MEMORY   2
-#define OGRERR_UNSUPPORTED_GEOMETRY_TYPE 3
-#define OGRERR_UNSUPPORTED_OPERATION 4
-#define OGRERR_CORRUPT_DATA        5
-#define OGRERR_FAILURE             6
-#define OGRERR_UNSUPPORTED_SRS     7
-#define OGRERR_INVALID_HANDLE      8
-#define OGRERR_NON_EXISTING_FEATURE 9   /* added in GDAL 2.0 */
+#define OGRERR_NONE                0       /**< Success */
+#define OGRERR_NOT_ENOUGH_DATA     1       /**< Not enough data to deserialize */
+#define OGRERR_NOT_ENOUGH_MEMORY   2       /**< Not enough memory */
+#define OGRERR_UNSUPPORTED_GEOMETRY_TYPE 3 /**< Unsupported geometry type */
+#define OGRERR_UNSUPPORTED_OPERATION 4     /**< Unsupported operation */
+#define OGRERR_CORRUPT_DATA        5       /**< Corrupt data */
+#define OGRERR_FAILURE             6       /**< Failure */
+#define OGRERR_UNSUPPORTED_SRS     7       /**< Unsupported SRS */
+#define OGRERR_INVALID_HANDLE      8       /**< Invalid handle */
+#define OGRERR_NON_EXISTING_FEATURE 9      /**< Non existing feature. Added in GDAL 2.0 */
 
 #endif
 
+/** Type for a OGR boolean */
 typedef int     OGRBoolean;
 
 /* -------------------------------------------------------------------- */
@@ -360,7 +372,7 @@ typedef enum
                                *   ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
     wkbTIN = 16,              /**< a PolyhedralSurface consisting only of Triangle patches
                                *    ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
-    wkbTriangle = 17,         /** < a Triangle. ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTriangle = 17,         /**< a Triangle. ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
 
     wkbNone = 100,          /**< non-standard, for pure attribute records */
     wkbLinearRing = 101,    /**< non-standard, just for createGeometry() */
@@ -412,6 +424,16 @@ typedef enum
     wkbTINZM = 3016,                /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
     wkbTriangleZM = 3017,           /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
 
+#if defined(DOXYGEN_SKIP)
+    // Sphinx doesn't like 0x8000000x constants
+    wkbPoint25D = -2147483647, /**< 2.5D extension as per 99-402 */
+    wkbLineString25D = -2147483646, /**< 2.5D extension as per 99-402 */
+    wkbPolygon25D = -2147483645, /**< 2.5D extension as per 99-402 */
+    wkbMultiPoint25D = -2147483644, /**< 2.5D extension as per 99-402 */
+    wkbMultiLineString25D = -2147483643, /**< 2.5D extension as per 99-402 */
+    wkbMultiPolygon25D = -2147483642, /**< 2.5D extension as per 99-402 */
+    wkbGeometryCollection25D = -2147483641 /**< 2.5D extension as per 99-402 */
+#else
     wkbPoint25D = 0x80000001, /**< 2.5D extension as per 99-402 */
     wkbLineString25D = 0x80000002, /**< 2.5D extension as per 99-402 */
     wkbPolygon25D = 0x80000003, /**< 2.5D extension as per 99-402 */
@@ -419,7 +441,7 @@ typedef enum
     wkbMultiLineString25D = 0x80000005, /**< 2.5D extension as per 99-402 */
     wkbMultiPolygon25D = 0x80000006, /**< 2.5D extension as per 99-402 */
     wkbGeometryCollection25D = 0x80000007 /**< 2.5D extension as per 99-402 */
-
+#endif
 } OGRwkbGeometryType;
 
 /**
@@ -443,14 +465,18 @@ typedef enum
     wkbVariantPostGIS1 /**< PostGIS 1.X has different codes for CurvePolygon, MultiCurve and MultiSurface */
 } OGRwkbVariant;
 
-
-/** @deprecated in GDAL 2.0. Use wkbHasZ() or wkbSetZ() instead */
 #ifndef GDAL_COMPILATION
+/** @deprecated in GDAL 2.0. Use wkbHasZ() or wkbSetZ() instead */
 #define wkb25DBit 0x80000000
 #endif
 
+#ifndef __cplusplus
 /** Return the 2D geometry type corresponding to the specified geometry type */
 #define wkbFlatten(x)  OGR_GT_Flatten((OGRwkbGeometryType)(x))
+#else
+/** Return the 2D geometry type corresponding to the specified geometry type */
+#define wkbFlatten(x)  OGR_GT_Flatten(static_cast<OGRwkbGeometryType>(x))
+#endif
 
 /** Return if the geometry type is a 3D geometry type
   * @since GDAL 2.0
@@ -472,7 +498,9 @@ typedef enum
   */
 #define wkbSetM(x)     OGR_GT_SetM(x)
 
+#ifndef DOXYGEN_SKIP
 #define ogrZMarker 0x21125711
+#endif
 
 const char CPL_DLL * OGRGeometryTypeToName( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGRMergeGeometryTypes( OGRwkbGeometryType eMain,
@@ -495,24 +523,29 @@ OGRwkbGeometryType CPL_DLL OGR_GT_GetCollection( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGR_GT_GetCurve( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGR_GT_GetLinear( OGRwkbGeometryType eType );
 
+/** Enumeration to describe byte order */
 typedef enum
 {
-    wkbXDR = 0,         /* MSB/Sun/Motoroloa: Most Significant Byte First   */
-    wkbNDR = 1          /* LSB/Intel/Vax: Least Significant Byte First      */
+    wkbXDR = 0,         /**< MSB/Sun/Motoroloa: Most Significant Byte First   */
+    wkbNDR = 1          /**< LSB/Intel/Vax: Least Significant Byte First      */
 } OGRwkbByteOrder;
 
+#ifndef DOXYGEN_SKIP
+
 #ifndef NO_HACK_FOR_IBM_DB2_V72
 #  define HACK_FOR_IBM_DB2_V72
 #endif
 
 #ifdef HACK_FOR_IBM_DB2_V72
 #  define DB2_V72_FIX_BYTE_ORDER(x) ((((x) & 0x31) == (x)) ? ((x) & 0x1) : (x))
-#  define DB2_V72_UNFIX_BYTE_ORDER(x) ((unsigned char) (OGRGeometry::bGenerate_DB2_V72_BYTE_ORDER ? ((x) | 0x30) : (x)))
+#  define DB2_V72_UNFIX_BYTE_ORDER(x) CPL_STATIC_CAST(unsigned char, OGRGeometry::bGenerate_DB2_V72_BYTE_ORDER ? ((x) | 0x30) : (x))
 #else
 #  define DB2_V72_FIX_BYTE_ORDER(x) (x)
 #  define DB2_V72_UNFIX_BYTE_ORDER(x) (x)
 #endif
 
+#endif /* #ifndef DOXYGEN_SKIP */
+
 /** Alter field name.
  * Used by OGR_L_AlterFieldDefn().
  */
@@ -540,11 +573,17 @@ typedef enum
  */
 #define ALTER_DEFAULT_FLAG         0x10
 
+/** Alter field UNIQUE constraint.
+ * Used by OGR_L_AlterFieldDefn().
+ * @since GDAL 3.2
+ */
+#define ALTER_UNIQUE_FLAG         0x20
+
+
 /** Alter all parameters of field definition.
  * Used by OGR_L_AlterFieldDefn().
  */
-#define ALTER_ALL_FLAG             (ALTER_NAME_FLAG | ALTER_TYPE_FLAG | ALTER_WIDTH_PRECISION_FLAG | ALTER_NULLABLE_FLAG | ALTER_DEFAULT_FLAG)
-
+#define ALTER_ALL_FLAG             (ALTER_NAME_FLAG | ALTER_TYPE_FLAG | ALTER_WIDTH_PRECISION_FLAG | ALTER_NULLABLE_FLAG | ALTER_DEFAULT_FLAG | ALTER_UNIQUE_FLAG)
 
 /** Validate that fields respect not-null constraints.
  * Used by OGR_F_Validate().
@@ -634,7 +673,11 @@ typedef enum
                                                         OFSTInt16 = 2,
     /** Single precision (32 bit) floating point. Only valid for OFTReal and OFTRealList. */
                                                         OFSTFloat32 = 3,
-                                                        OFSTMaxSubType = 3
+    /** JSON content. Only valid for OFTString.
+     * @since GDAL 2.4
+     */
+                                                        OFSTJSON = 4,
+                                                        OFSTMaxSubType = 4
 } OGRFieldSubType;
 
 /**
@@ -648,9 +691,32 @@ typedef enum
     OJRight = 2
 } OGRJustification;
 
+/** Special value for a unset FID */
 #define OGRNullFID            -1
+
+/* Special value for an unknown field type. This should only be used
+ * while reading a file. At the end of file any unknown types should
+ * be set to OFTString.
+*/
+/*! @cond Doxygen_Suppress */
+#define OGRUnknownType        static_cast<OGRFieldType>(-1)
+/*! @endcond */
+
+/** Special value set in OGRField.Set.nMarker1, nMarker2 and nMarker3 for
+ *  a unset field.
+ *  Direct use of this value is strongly discouraged.
+ *  Use OGR_RawField_SetUnset() or OGR_RawField_IsUnset() instead.
+ */
 #define OGRUnsetMarker        -21121
 
+/** Special value set in OGRField.Set.nMarker1, nMarker2 and nMarker3 for
+ *  a null field.
+ *  Direct use of this value is strongly discouraged.
+ *  Use OGR_RawField_SetNull() or OGR_RawField_IsNull() instead.
+ *  @since GDAL 2.2
+ */
+#define OGRNullMarker         -21122
+
 /************************************************************************/
 /*                               OGRField                               */
 /************************************************************************/
@@ -660,6 +726,7 @@ typedef enum
  */
 
 typedef union {
+/*! @cond Doxygen_Suppress */
     int         Integer;
     GIntBig     Integer64;
     double      Real;
@@ -693,6 +760,7 @@ typedef union {
     struct {
         int     nMarker1;
         int     nMarker2;
+        int     nMarker3;
     } Set;
 
     struct {
@@ -706,9 +774,19 @@ typedef union {
         GByte   Reserved; /* must be set to 0 */
         float   Second; /* with millisecond accuracy. at the end of the structure, so as to keep it 12 bytes on 32 bit */
     } Date;
+/*! @endcond */
 } OGRField;
 
-#define OGR_GET_MS(floatingpoint_sec)   (int)(((floatingpoint_sec) - (int)(floatingpoint_sec)) * 1000 + 0.5)
+#ifdef __cplusplus
+/** Return the number of milliseconds from a datetime with decimal seconds */
+inline int OGR_GET_MS(float fSec) {
+  if( CPLIsNan(fSec) ) return 0;
+  if( fSec >= 999 ) return 999;
+  if( fSec <= 0 ) return 0;
+  const float fValue = (fSec - static_cast<int>(fSec)) * 1000 + 0.5f;
+  return static_cast<int>(fValue);
+}
+#endif  // __cplusplus
 
 int CPL_DLL OGRParseDate( const char *pszInput, OGRField *psOutput,
                           int nOptions );
@@ -716,35 +794,38 @@ int CPL_DLL OGRParseDate( const char *pszInput, OGRField *psOutput,
 /* -------------------------------------------------------------------- */
 /*      Constants from ogrsf_frmts.h for capabilities.                  */
 /* -------------------------------------------------------------------- */
-#define OLCRandomRead          "RandomRead"
-#define OLCSequentialWrite     "SequentialWrite"
-#define OLCRandomWrite         "RandomWrite"
-#define OLCFastSpatialFilter   "FastSpatialFilter"
-#define OLCFastFeatureCount    "FastFeatureCount"
-#define OLCFastGetExtent       "FastGetExtent"
-#define OLCCreateField         "CreateField"
-#define OLCDeleteField         "DeleteField"
-#define OLCReorderFields       "ReorderFields"
-#define OLCAlterFieldDefn      "AlterFieldDefn"
-#define OLCTransactions        "Transactions"
-#define OLCDeleteFeature       "DeleteFeature"
-#define OLCFastSetNextByIndex  "FastSetNextByIndex"
-#define OLCStringsAsUTF8       "StringsAsUTF8"
-#define OLCIgnoreFields        "IgnoreFields"
-#define OLCCreateGeomField     "CreateGeomField"
-#define OLCCurveGeometries     "CurveGeometries"
-#define OLCMeasuredGeometries  "MeasuredGeometries"
+#define OLCRandomRead          "RandomRead"         /**< Layer capability for random read */
+#define OLCSequentialWrite     "SequentialWrite"    /**< Layer capability for sequential write */
+#define OLCRandomWrite         "RandomWrite"        /**< Layer capability for random write */
+#define OLCFastSpatialFilter   "FastSpatialFilter"  /**< Layer capability for fast spatial filter */
+#define OLCFastFeatureCount    "FastFeatureCount"   /**< Layer capability for fast feature count retrieval */
+#define OLCFastGetExtent       "FastGetExtent"      /**< Layer capability for fast extent retrieval */
+#define OLCCreateField         "CreateField"        /**< Layer capability for field creation */
+#define OLCDeleteField         "DeleteField"        /**< Layer capability for field deletion */
+#define OLCReorderFields       "ReorderFields"      /**< Layer capability for field reordering */
+#define OLCAlterFieldDefn      "AlterFieldDefn"     /**< Layer capability for field alteration */
+#define OLCTransactions        "Transactions"       /**< Layer capability for transactions */
+#define OLCDeleteFeature       "DeleteFeature"      /**< Layer capability for feature deletion */
+#define OLCFastSetNextByIndex  "FastSetNextByIndex" /**< Layer capability for setting next feature index */
+#define OLCStringsAsUTF8       "StringsAsUTF8"      /**< Layer capability for strings returned with UTF-8 encoding */
+#define OLCIgnoreFields        "IgnoreFields"       /**< Layer capability for field ignoring */
+#define OLCCreateGeomField     "CreateGeomField"    /**< Layer capability for geometry field creation */
+#define OLCCurveGeometries     "CurveGeometries"    /**< Layer capability for curve geometries support */
+#define OLCMeasuredGeometries  "MeasuredGeometries" /**< Layer capability for measured geometries support */
 
-#define ODsCCreateLayer        "CreateLayer"
-#define ODsCDeleteLayer        "DeleteLayer"
-#define ODsCCreateGeomFieldAfterCreateLayer   "CreateGeomFieldAfterCreateLayer"
-#define ODsCCurveGeometries    "CurveGeometries"
-#define ODsCTransactions       "Transactions"
-#define ODsCEmulatedTransactions "EmulatedTransactions"
-#define ODsCMeasuredGeometries "MeasuredGeometries"
+#define ODsCCreateLayer        "CreateLayer"        /**< Dataset capability for layer creation */
+#define ODsCDeleteLayer        "DeleteLayer"        /**< Dataset capability for layer deletion */
+#define ODsCCreateGeomFieldAfterCreateLayer   "CreateGeomFieldAfterCreateLayer" /**< Dataset capability for geometry field creation support */
+#define ODsCCurveGeometries    "CurveGeometries"    /**< Dataset capability for curve geometries support */
+#define ODsCTransactions       "Transactions"       /**< Dataset capability for dataset transcations */
+#define ODsCEmulatedTransactions "EmulatedTransactions" /**< Dataset capability for emulated dataset transactions */
+#define ODsCMeasuredGeometries "MeasuredGeometries"     /**< Dataset capability for measured geometries support */
+#define ODsCRandomLayerRead     "RandomLayerRead"   /**< Dataset capability for GetNextFeature() returning features from random layers */
+/* Note the unfortunate trailing space at the end of the string */
+#define ODsCRandomLayerWrite    "RandomLayerWrite " /**< Dataset capability for supporting CreateFeature on layer in random order */
 
-#define ODrCCreateDataSource   "CreateDataSource"
-#define ODrCDeleteDataSource   "DeleteDataSource"
+#define ODrCCreateDataSource   "CreateDataSource"   /**< Driver capability for datasource creation */
+#define ODrCDeleteDataSource   "DeleteDataSource"   /**< Driver capability for datasource deletion */
 
 /* -------------------------------------------------------------------- */
 /*      Layer metadata items.                                           */
@@ -765,12 +846,12 @@ int CPL_DLL OGRParseDate( const char *pszInput, OGRField *psOutput,
 
 typedef enum ogr_style_tool_class_id
 {
-    OGRSTCNone   = 0,
-    OGRSTCPen    = 1,
-    OGRSTCBrush  = 2,
-    OGRSTCSymbol = 3,
-    OGRSTCLabel  = 4,
-    OGRSTCVector = 5
+    OGRSTCNone   = 0, /**< None */
+    OGRSTCPen    = 1, /**< Pen */
+    OGRSTCBrush  = 2, /**< Brush */
+    OGRSTCSymbol = 3, /**< Symbol */
+    OGRSTCLabel  = 4, /**< Label */
+    OGRSTCVector = 5  /**< Vector */
 } OGRSTClassId;
 
 /**
@@ -778,12 +859,12 @@ typedef enum ogr_style_tool_class_id
  */
 typedef enum ogr_style_tool_units_id
 {
-    OGRSTUGround = 0,
-    OGRSTUPixel  = 1,
-    OGRSTUPoints = 2,
-    OGRSTUMM     = 3,
-    OGRSTUCM     = 4,
-    OGRSTUInches = 5
+    OGRSTUGround = 0, /**< Ground unit */
+    OGRSTUPixel  = 1, /**< Pixel */
+    OGRSTUPoints = 2, /**< Points */
+    OGRSTUMM     = 3, /**< Millimeter */
+    OGRSTUCM     = 4, /**< Centimeter */
+    OGRSTUInches = 5  /**< Inch */
 } OGRSTUnitId;
 
 /**
@@ -791,16 +872,17 @@ typedef enum ogr_style_tool_units_id
  */
 typedef enum ogr_style_tool_param_pen_id
 {
-    OGRSTPenColor       = 0,
-    OGRSTPenWidth       = 1,
-    OGRSTPenPattern     = 2,
-    OGRSTPenId          = 3,
-    OGRSTPenPerOffset   = 4,
-    OGRSTPenCap         = 5,
-    OGRSTPenJoin        = 6,
-    OGRSTPenPriority    = 7,
+    OGRSTPenColor       = 0, /**< Color */
+    OGRSTPenWidth       = 1, /**< Width */
+    OGRSTPenPattern     = 2, /**< Pattern */
+    OGRSTPenId          = 3, /**< Id */
+    OGRSTPenPerOffset   = 4, /**< Perpendicular offset */
+    OGRSTPenCap         = 5, /**< Cap */
+    OGRSTPenJoin        = 6, /**< Join */
+    OGRSTPenPriority    = 7, /**< Priority */
+#ifndef DOXYGEN_SKIP
     OGRSTPenLast        = 8
-
+#endif
 } OGRSTPenParam;
 
 /**
@@ -808,38 +890,40 @@ typedef enum ogr_style_tool_param_pen_id
  */
 typedef enum ogr_style_tool_param_brush_id
 {
-    OGRSTBrushFColor    = 0,
-    OGRSTBrushBColor    = 1,
-    OGRSTBrushId        = 2,
-    OGRSTBrushAngle     = 3,
-    OGRSTBrushSize      = 4,
-    OGRSTBrushDx        = 5,
-    OGRSTBrushDy        = 6,
-    OGRSTBrushPriority  = 7,
+    OGRSTBrushFColor    = 0, /**< Foreground color */
+    OGRSTBrushBColor    = 1, /**< Background color */
+    OGRSTBrushId        = 2, /**< Id */
+    OGRSTBrushAngle     = 3, /**< Angle */
+    OGRSTBrushSize      = 4, /**< Size */
+    OGRSTBrushDx        = 5, /**< Dx */
+    OGRSTBrushDy        = 6, /**< Dy */
+    OGRSTBrushPriority  = 7, /**< Priority */
+#ifndef DOXYGEN_SKIP
     OGRSTBrushLast      = 8
+#endif
 
 } OGRSTBrushParam;
 
-
 /**
  * List of parameters for use with OGRStyleSymbol.
  */
 typedef enum ogr_style_tool_param_symbol_id
 {
-    OGRSTSymbolId       = 0,
-    OGRSTSymbolAngle    = 1,
-    OGRSTSymbolColor    = 2,
-    OGRSTSymbolSize     = 3,
-    OGRSTSymbolDx       = 4,
-    OGRSTSymbolDy       = 5,
-    OGRSTSymbolStep     = 6,
-    OGRSTSymbolPerp     = 7,
-    OGRSTSymbolOffset   = 8,
-    OGRSTSymbolPriority = 9,
-    OGRSTSymbolFontName = 10,
-    OGRSTSymbolOColor   = 11,
+    OGRSTSymbolId       = 0, /**< Id */
+    OGRSTSymbolAngle    = 1, /**< Angle */
+    OGRSTSymbolColor    = 2, /**< Color */
+    OGRSTSymbolSize     = 3, /**< Size */
+    OGRSTSymbolDx       = 4, /**< Dx */
+    OGRSTSymbolDy       = 5, /**< Dy */
+    OGRSTSymbolStep     = 6, /**< Step */
+    OGRSTSymbolPerp     = 7, /**< Perpendicular */
+    OGRSTSymbolOffset   = 8, /**< Offset */
+    OGRSTSymbolPriority = 9, /**< Priority */
+    OGRSTSymbolFontName = 10, /**< Font name */
+    OGRSTSymbolOColor   = 11, /**< Outline color */
+#ifndef DOXYGEN_SKIP
     OGRSTSymbolLast     = 12
-
+#endif
 } OGRSTSymbolParam;
 
 /**
@@ -847,35 +931,38 @@ typedef enum ogr_style_tool_param_symbol_id
  */
 typedef enum ogr_style_tool_param_label_id
 {
-    OGRSTLabelFontName  = 0,
-    OGRSTLabelSize      = 1,
-    OGRSTLabelTextString = 2,
-    OGRSTLabelAngle     = 3,
-    OGRSTLabelFColor    = 4,
-    OGRSTLabelBColor    = 5,
-    OGRSTLabelPlacement = 6,
-    OGRSTLabelAnchor    = 7,
-    OGRSTLabelDx        = 8,
-    OGRSTLabelDy        = 9,
-    OGRSTLabelPerp      = 10,
-    OGRSTLabelBold      = 11,
-    OGRSTLabelItalic    = 12,
-    OGRSTLabelUnderline = 13,
-    OGRSTLabelPriority  = 14,
-    OGRSTLabelStrikeout = 15,
-    OGRSTLabelStretch   = 16,
-    OGRSTLabelAdjHor    = 17,
-    OGRSTLabelAdjVert   = 18,
-    OGRSTLabelHColor    = 19,
-    OGRSTLabelOColor    = 20,
+    OGRSTLabelFontName  = 0, /**< Font name */
+    OGRSTLabelSize      = 1, /**< Size */
+    OGRSTLabelTextString = 2, /**< Text string */
+    OGRSTLabelAngle     = 3, /**< Angle */
+    OGRSTLabelFColor    = 4, /**< Foreground color */
+    OGRSTLabelBColor    = 5, /**< Background color */
+    OGRSTLabelPlacement = 6, /**< Placement */
+    OGRSTLabelAnchor    = 7, /**< Anchor */
+    OGRSTLabelDx        = 8, /**< Dx */
+    OGRSTLabelDy        = 9, /**< Dy */
+    OGRSTLabelPerp      = 10, /**< Perpendicular */
+    OGRSTLabelBold      = 11, /**< Bold */
+    OGRSTLabelItalic    = 12, /**< Italic */
+    OGRSTLabelUnderline = 13, /**< Underline */
+    OGRSTLabelPriority  = 14, /**< Priority */
+    OGRSTLabelStrikeout = 15, /**< Strike out */
+    OGRSTLabelStretch   = 16, /**< Stretch */
+    OGRSTLabelAdjHor    = 17, /**< OBSOLETE; do not use */
+    OGRSTLabelAdjVert   = 18, /**< OBSOLETE; do not use */
+    OGRSTLabelHColor    = 19, /**< Highlight color */
+    OGRSTLabelOColor    = 20, /**< Outline color */
+#ifndef DOXYGEN_SKIP
     OGRSTLabelLast      = 21
-
+#endif
 } OGRSTLabelParam;
 
 /* ------------------------------------------------------------------- */
 /*                        Version checking                             */
 /* -------------------------------------------------------------------- */
 
+#ifndef DOXYGEN_SKIP
+
 /* Note to developers : please keep this section in sync with gdal.h */
 
 #ifndef GDAL_VERSION_INFO_DEFINED
@@ -905,6 +992,8 @@ int CPL_DLL CPL_STDCALL GDALCheckVersion( int nVersionMajor, int nVersionMinor,
 
 #endif
 
+#endif /* #ifndef DOXYGEN_SKIP */
+
 CPL_C_END
 
 #endif /* ndef OGR_CORE_H_INCLUDED */
